# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from latex2pydata import loads as latex2pydata_loads



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

from .command_clean import clean_temp
from .messages import Messages
from .restricted import cwd_path, input_tempfiledir_path_str, tempfiledir_path, RestrictedPath
from .version import __version_info__




def config(*, md5: str, timestamp: str, messages: Messages, data: dict[str, str] | None):
    config_file_name = f'_{md5}.config.minted'

    clean_temp(md5=md5)

    config_lines = []
    minted_executable_version = f'{__version_info__.major}.{__version_info__.minor}.{__version_info__.micro}'
    config_lines.append(rf'\gdef\minted@executable@version{{{minted_executable_version}}}%')
    config_lines.append(rf'\gdef\minted@executable@timestamp{{{timestamp}}}%')
    if data is not None:
        tex_timestamp: str = data['timestamp']
        config_lines.append(rf'\gdef\minted@config@timestamp{{{tex_timestamp}}}%')
        tex_cachepath: str
        tex_cachedir: str = data['cachedir']
        if RestrictedPath(tex_cachedir).is_absolute() or not input_tempfiledir_path_str:
            tex_cachepath = tex_cachedir
        else:
            if tex_cachedir.startswith('./'):
                tex_cachedir = tex_cachedir[2:]
            tex_cachepath = f'{input_tempfiledir_path_str}{tex_cachedir}'
        if not tex_cachepath.endswith('/'):
            tex_cachepath += '/'
        config_lines.append(rf'\gdef\minted@config@cachepath{{{tex_cachepath}}}%')

    try:
        (tempfiledir_path / config_file_name).write_text('\n'.join(config_lines))
    except PermissionError:
        messages.append_error('Insufficient permission to write minted config file')
    else:
        return

    if data is None and tempfiledir_path is not cwd_path:
        try:
            (cwd_path / config_file_name).write_text('\n'.join(config_lines))
        except PermissionError:
            messages.append_error('Insufficient permission to write minted config file')



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

import hashlib
import re
import sys
import textwrap
from typing import Any
from pygments import highlight as pygments_highlight
from pygments.formatters import LatexFormatter
from pygments.lexer import Lexer
from pygments.lexers import get_lexer_by_name
from pygments.util import ClassNotFound
from .err import (
    PathSecurityError,
    SubprocessExecutableNotFoundError, SubprocessExecutablePathSecurityError,
    MissingLatexMintedRCError,
    CustomLexerError, CustomLexerSecurityError,
)
from .messages import Messages
from .restricted import load_custom_lexer, RestrictedPath, restricted_run




# Sets and dicts for checking Python-related options and converting values to
# Python types
bool_keys: set[str] = set([
    'autogobble',
    'codetagify',
    'funcnamehighlighting',
    'mathescape',
    'python3',
    'rangeregexdotall',
    'rangeregexmultiline',
    'startinline',
    'stripall',
    'stripnl',
    'texcl',
    'texcomments',
])
nonnegative_int_re = re.compile('[1-9][0-9]*|0')
nonnegative_int_or_none_keys: set[str] = set([
    'gobble',
])
positive_int_re = re.compile('[1-9][0-9]*')
positive_int_keys: set[str] = set([
    'rangeregexmatchnumber',
])
other_keys_value_sets: dict[str, set[str]] = {
    'keywordcase': set(['lower', 'upper' or 'capitalize']),
}
other_keys_unchecked_str_value: set[str] = set([
    'commandprefix',
    'encoding',
    'escapeinside',
    'lexer',
    'rangestartstring',
    'rangestartafterstring',
    'rangestopstring',
    'rangestopbeforestring',
    'rangeregex',
])
all_keys = bool_keys | nonnegative_int_or_none_keys | positive_int_keys | set(other_keys_value_sets) | other_keys_unchecked_str_value

# Key for manipulating code within Python
code_keys = set([
    'autogobble',
    'gobble',
    'rangestartstring',
    'rangestartafterstring',
    'rangestopstring',
    'rangestopbeforestring',
    'rangeregex',
    'rangeregexmatchnumber',
    'rangeregexdotall',
    'rangeregexmultiline',
])
# Categorize Pygments options into lexer, filter, or formatter
lexer_keys: set[str] = set([
    'funcnamehighlighting',
    'python3',
    'startinline',
    'stripall',
    'stripnl',
])
filter_keys_no_options: set[str] = set([
    'codetagify',
])
filter_keys_one_option: dict[str, str] = {
    'keywordcase': 'case',
}
filter_keys = filter_keys_no_options | set(filter_keys_one_option)
formatter_keys: set[str] = set([
    'commandprefix',
    'escapeinside',
    'mathescape',
    'styleprefix',
    'texcl',
    'texcomments',
])
pygments_keys = lexer_keys | filter_keys | formatter_keys


def process_highlight_data(*, messages: Messages, data: dict[str, Any]) -> tuple[dict[str, Any], ...] | None:
    minted_opts: dict[str, str] = {k: v for k, v in data.items() if k != 'pyopt'}
    py_opts = {}
    code_opts = {}
    lexer_opts = {}
    filter_opts = {}
    formatter_opts = {}

    for k, v in data['pyopt'].items():
        if k in lexer_keys:
            current_opts = lexer_opts
        elif k in filter_keys:
            current_opts = filter_opts
        elif k in formatter_keys:
            current_opts = formatter_opts
        elif k in code_keys:
            current_opts = code_opts
        elif k in all_keys:
            current_opts = py_opts
        else:
            messages.append_error(rf'Key "{k}" is unknown and will be ignored')
            continue

        if k in bool_keys:
            if v in ('true', 'false'):
                current_opts[k] = v == 'true'
            else:
                messages.append_error(rf'Key "{k}" has invalid value \detokenize{{"{v}"}} (expected "true" or "false")')
        elif k in nonnegative_int_or_none_keys:
            if v == 'none':
                current_opts[k] = 0
            elif nonnegative_int_re.fullmatch(v):
                current_opts[k] = int(v)
            else:
                messages.append_error(rf'Key "{k}" has invalid value \detokenize{{"{v}"}} (expected non-negative integer or "none")')
        elif k in positive_int_keys:
            if positive_int_re.fullmatch(v):
                current_opts[k] = int(v)
            else:
                messages.append_error(rf'Key "{k}" has invalid value \detokenize{{"{v}"}} (expected positive integer)')
        elif k in other_keys_value_sets:
            if v in other_keys_value_sets[k]:
                current_opts[k] = v
            else:
                valid_options = ', '.join(f'"{opt}"' for opt in other_keys_value_sets[k])
                messages.append_error(rf'Key "{k}" has invalid value \detokenize{{"{v}"}} (expected {valid_options})')
        elif k in other_keys_unchecked_str_value:
            current_opts[k] = v
        else:
            messages.append_error(rf'Unknown key "{k}"')

    # Additional data processing
    if 'inputfilemdfivesum' in minted_opts:
        minted_opts['inputfilemdfivesum'] = minted_opts['inputfilemdfivesum'].lower()

    if messages.has_errors():
        return None
    return minted_opts, py_opts, code_opts, lexer_opts, filter_opts, formatter_opts




def load_input_file(*, messages: Messages, input_file: str, mdfivesum: str, encoding: str) -> str | None:
    input_file_path = RestrictedPath(input_file)
    try:
        input_file_bytes = input_file_path.read_bytes()
        hasher = hashlib.md5()
        hasher.update(input_file_bytes)
        if hasher.hexdigest() != mdfivesum:
            raise FileNotFoundError
        code_bytes = input_file_bytes
    except PermissionError:
        messages.append_error(rf'Insufficient permission to open file \detokenize{{"{input_file}"}}')
        return None
    except PathSecurityError:
        messages.append_error(rf'Cannot read file in prohibited location: \detokenize{{"{input_file}"}}')
        return None
    except FileNotFoundError:
        try:
            kpsewhich_proc = restricted_run(['kpsewhich', input_file])
        except SubprocessExecutableNotFoundError:
            messages.append_error(rf'Cannot locate file \detokenize{{"{input_file}"}} (kpsewhich unavailable)')
            return None
        except SubprocessExecutablePathSecurityError:
            messages.append_error(
                rf'Cannot use kpsewhich to locate file \detokenize{{"{input_file}"}} '
                r'when kpsewhich is located under the current directory, \detokenize{TEXMFOUTPUT}, '
                r'or \detokenize{TEXMF_OUTPUT_DIRECTORY}, or when these locations are under the same '
                r'directory as kpsewhich'
            )
            return None
        if kpsewhich_proc.returncode != 0:
            sys.stderr.buffer.write(kpsewhich_proc.stderr)
            messages.append_error(rf'Cannot locate file \detokenize{{"{input_file}"}} (kpsewhich failed)')
            return None
        try:
            kpsewhich_input_file_path = RestrictedPath(kpsewhich_proc.stdout.decode(sys.stdout.encoding).strip())
        except UnicodeDecodeError:
            messages.append_error(rf'Cannot locate file \detokenize{{"{input_file}"}} (could not decode kpsewhich output)')
            return None
        try:
            kpsewhich_input_file_bytes = kpsewhich_input_file_path.read_bytes()
        except FileNotFoundError:
            messages.append_error(rf'Cannot locate file \detokenize{{"{input_file}"}}')
            return None
        except PermissionError:
            messages.append_error(rf'Insufficient permission to open file \detokenize{{"{input_file}"}}')
            return None
        except PathSecurityError:
            messages.append_error(rf'Cannot read file in prohibited location: \detokenize{{"{input_file}"}}')
            return None
        hasher = hashlib.md5()
        hasher.update(kpsewhich_input_file_bytes)
        if hasher.hexdigest() != mdfivesum:
            messages.append_error(
                rf'Cannot find the correct input file \detokenize{{"{input_file}"}}; '
                'there may be multiple files with the same name, or the file may have been modified during compilation'
            )
            return None
        code_bytes = kpsewhich_input_file_bytes
    try:
        code = code_bytes.decode(encoding=encoding)
    except UnicodeDecodeError as e:
        messages.append_error(rf'Cannot decode file \detokenize{{"{input_file}"}} with encoding=\detokenize{{"{encoding}"}}')
        return None
    return code




def preprocess_code(code: str, *, messages: Messages,
                    autogobble: bool, gobble: int,
                    rangestartstring: str, rangestartafterstring: str, rangestopstring: str, rangestopbeforestring: str,
                    rangeregex: str, rangeregexmatchnumber: int, rangeregexdotall: bool, rangeregexmultiline: bool) -> str | None:
    if rangeregex and any([rangestartstring, rangestartafterstring, rangestopstring, rangestopbeforestring]):
        messages.append_error('Cannot use "rangeregex" at the same time as range string options')
        return
    if rangestartstring and rangestartafterstring:
        messages.append_error('Cannot use "rangestartstring" and "rangestartafterstring" at the same time')
        return
    if rangestopstring and rangestopbeforestring:
        messages.append_error('Cannot use "rangestopstring" and "rangestopbeforestring" at the same time')
        return

    if rangeregex:
        flags = re.NOFLAG
        if rangeregexdotall:
            flags |= re.DOTALL
        if rangeregexmultiline:
            flags |= re.MULTILINE
        try:
            pattern = re.compile(rangeregex, flags)
        except re.error as e:
            messages.append_error(
                rf'Failed to compile "rangeregex" regular expression (see \detokenize{{{messages.errlog_file_name}}} if it exists)'
            )
            messages.append_errlog(e)
            return
        regex_match: re.Match | None = None
        for n, match_n in enumerate(pattern.finditer(code), start=1):
            if n == rangeregexmatchnumber:
                regex_match = match_n
                break
        if not regex_match:
            messages.append_error('Failed to find match with regular expression "rangeregex"')
            return
        code = regex_match.group()

    if rangestartstring:
        index = code.find(rangestartstring)
        if index == -1:
            messages.append_error('Failed to find string for "rangestartstring"')
            return
        code = code[index:]
    if rangestartafterstring:
        index = code.find(rangestartafterstring)
        if index == -1:
            messages.append_error('Failed to find string for "rangestartafterstring"')
            return
        code = code[index+len(rangestartafterstring):]

    if rangestopstring:
        index = code.find(rangestopstring)
        if index == -1:
            messages.append_error('Failed to find string for "rangestopstring"')
            return
        code = code[:index+len(rangestopstring)]
    if rangestopbeforestring:
        index = code.find(rangestopbeforestring)
        if index == -1:
            messages.append_error('Failed to find string for "rangestopbeforestring"')
            return
        code = code[:index]

    if autogobble:
        code = textwrap.dedent(code)
    if gobble:
        code = ''.join(line[gobble:] or '\n' for line in code.splitlines(True))

    if not code.endswith('\n'):
        code += '\n'

    return code




def highlight(*, md5: str, timestamp: str, messages: Messages, data: dict[str, Any]) -> str | None:
    processed_data = process_highlight_data(messages=messages, data=data)
    if processed_data is None:
        return

    minted_opts, py_opts, code_opts, lexer_opts, filter_opts, formatter_opts = processed_data

    if 'code' in minted_opts:
        code = minted_opts['code']
    else:
        code = load_input_file(messages=messages, input_file=minted_opts['inputfilepath'],
                               mdfivesum=minted_opts['inputfilemdfivesum'], encoding=py_opts['encoding'])
        if code is None:
            return

    code = preprocess_code(code, messages=messages, **code_opts)
    if code is None:
        return

    pygments_lexer: Lexer
    try:
        pygments_lexer = get_lexer_by_name(py_opts['lexer'], **lexer_opts)
    except ClassNotFound:
        if not py_opts['lexer'].endswith('.py') and '.py:' not in py_opts['lexer']:
            messages.append_error(rf'''Pygments lexer \detokenize{{"{py_opts['lexer']}"}} is unknown''')
            return
        try:
            pygments_lexer_class = load_custom_lexer(py_opts['lexer'])
        except (MissingLatexMintedRCError, CustomLexerError, CustomLexerSecurityError) as e:
            messages.append_error(rf'\detokenize{{{str(e)}}}')
            return
        except Exception as e:
            messages.append_error(
                rf'''Failed to load custom lexer \detokenize{{"{py_opts['lexer']}"}}; see \detokenize{{{messages.errlog_file_name}}} if it exists''')
            messages.append_errlog(e)
            return
        pygments_lexer = pygments_lexer_class(**lexer_opts)

    for filter_name in filter_keys_no_options:
        if filter_opts[filter_name]:
            pygments_lexer.add_filter(filter_name)
    for filter_name, opt_name in filter_keys_one_option.items():
        if filter_opts[filter_name]:
            pygments_lexer.add_filter(filter_name, **{opt_name: filter_opts[filter_name]})

    pygments_formatter = LatexFormatter(**formatter_opts)

    highlighted = pygments_highlight(code, pygments_lexer, pygments_formatter)
    highlighted_path = RestrictedPath(data['cachepath']) / minted_opts['highlightfilename']
    try:
        highlighted_path.write_text(highlighted)
    except PermissionError:
        messages.append_error(r'Insufficient permission to write highlighted code')
        return
    except PathSecurityError:
        messages.append_error(
            r'Cannot write highlighted code outside working directory, \detokenize{TEXMFOUTPUT}, and \detokenize{TEXMF_OUTPUT_DIRECTORY}'
        )
        return
    return minted_opts['highlightfilename']



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

from datetime import date, timedelta
from .err import PathSecurityError
from .messages import Messages
from .restricted import cwd_path, json_loads, json_dumps, RestrictedPath, tempfiledir_path, texmfoutput_path




file_roles = ['config', 'data', 'errlog', 'highlight', 'message', 'style']




def clean_messages(*, md5: str):
    paths = [tempfiledir_path]
    if cwd_path is not tempfiledir_path:
        paths.append(cwd_path)
    # Messages and error logs can be written to $TEXMFOUTPUT
    if texmfoutput_path is not None and texmfoutput_path not in paths:
        paths.append(texmfoutput_path)
    for path in paths:
        try:
            (path / f'_{md5}.message.minted').unlink(missing_ok=True)
        except (PermissionError, PathSecurityError):
            pass
        try:
            (path / f'_{md5}.errlog.minted').unlink(missing_ok=True)
        except (PermissionError, PathSecurityError):
            pass


def clean_temp(*, md5: str):
    paths = [tempfiledir_path]
    if cwd_path is not tempfiledir_path:
        paths.append(cwd_path)
    for path in paths:
        for role in file_roles:
            try:
                (path / f'_{md5}.{role}.minted').unlink(missing_ok=True)
            except (PermissionError, PathSecurityError):
                pass
    # Only messages and error logs can be written to $TEXMFOUTPUT
    if texmfoutput_path is not None and texmfoutput_path not in paths:
        try:
            (texmfoutput_path / f'_{md5}.message.minted').unlink(missing_ok=True)
        except (PermissionError, PathSecurityError):
            pass
        try:
            (texmfoutput_path / f'_{md5}.errlog.minted').unlink(missing_ok=True)
        except (PermissionError, PathSecurityError):
            pass


def clean_temp_except_errlog(*, md5: str):
    paths = [tempfiledir_path]
    if cwd_path is not tempfiledir_path:
        paths.append(cwd_path)
    for path in paths:
        for role in file_roles:
            if role == 'errlog':
                continue
            try:
                (path / f'_{md5}.{role}.minted').unlink(missing_ok=True)
            except (PermissionError, PathSecurityError):
                pass
    # Only messages and error logs can be written to $TEXMFOUTPUT
    if texmfoutput_path is not None and texmfoutput_path not in paths:
        try:
            (texmfoutput_path / f'_{md5}.message.minted').unlink(missing_ok=True)
        except (PermissionError, PathSecurityError):
            pass




def clean(*, md5: str, timestamp: str, messages: Messages, data: dict[str, str],
          additional_cache_file_names: list[str] | None = None):
    clean_temp_except_errlog(md5=md5)

    # Python < 3.11 requires `YYYY-MM-DD`
    timestamp_date = date.fromisoformat(f'{timestamp[:4]}-{timestamp[4:6]}-{timestamp[6:8]}')
    cache_path = RestrictedPath(data['cachepath'])
    current_index_name = f'_{md5}.index.minted'
    used_cache_files: set[str] = set()
    used_cache_files.add(current_index_name)
    if additional_cache_file_names is not None:
        used_cache_files.update(additional_cache_file_names)
    used_cache_files.update(data['cachefiles'])


    for index_path in cache_path.glob('*.index.minted'):
        if index_path.name == current_index_name:
            continue

        try:
            index_data = json_loads(index_path.read_bytes())
        except PathSecurityError:
            messages.append_error(
                rf'Cannot read file \detokenize{{"{index_path.name}"}} outside working directory, \detokenize{{TEXMFOUTPUT}}, and \detokenize{{TEXMF_OUTPUT_DIRECTORY}}'
            )
            continue
        except PermissionError:
            messages.append_error(rf'Insufficient permission to open file \detokenize{{"{index_path.name}"}}')
            continue

        # Python < 3.11 requires `YYYY-MM-DD`
        index_data_ts = index_data['timestamp']
        index_date = date.fromisoformat(f'{index_data_ts[:4]}-{index_data_ts[4:6]}-{index_data_ts[6:8]}')
        index_age = timestamp_date - index_date
        if index_age > timedelta(days=30):
            # Delete index files more than 30 days old
            try:
                index_path.unlink(missing_ok=True)
            except PathSecurityError:
                messages.append_error(
                    rf'Cannot delete file \detokenize{{"{index_data.name}"}} outside working directory, \detokenize{{TEXMFOUTPUT}}, and \detokenize{{TEXMF_OUTPUT_DIRECTORY}}'
                )
            except PermissionError:
                messages.append_error(rf'Insufficient permission to delete outdated cache file \detokenize{{"{index_path.name}"}}')
            continue

        used_cache_files.add(index_path.name)
        used_cache_files.update(index_data['cachefiles'])

    for minted_path in cache_path.glob('*.minted'):
        if minted_path.name not in used_cache_files:
            try:
                minted_path.unlink(missing_ok=True)
            except PathSecurityError:
                messages.append_error(
                    rf'Cannot delete file \detokenize{{"{minted_path.name}"}} outside working directory, \detokenize{{TEXMFOUTPUT}}, and \detokenize{{TEXMF_OUTPUT_DIRECTORY}}'
                )
            except PermissionError:
                messages.append_error(rf'Insufficient permission to delete unused cache file \detokenize{{"{minted_path.name}"}}')

    new_index_data = {
        'jobname': data['jobname'],
        'md5': md5,
        'timestamp': timestamp,
        'cachefiles': sorted(used_cache_files),
    }
    new_index_path = cache_path / current_index_name
    try:
        new_index_path.write_text(json_dumps(new_index_data, indent=2))
    except PathSecurityError:
        messages.append_error(
            rf'Cannot write file \detokenize{{"{new_index_path.name}"}} outside working directory, \detokenize{{TEXMFOUTPUT}}, and \detokenize{{TEXMF_OUTPUT_DIRECTORY}}'
        )
    except PermissionError:
        messages.append_error(rf'Insufficient permission to write file \detokenize{{"{new_index_path.name}"}}')



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

from .restricted import os_environ
# ruff: noqa: E402
os_environ['PYDEVD_DISABLE_FILE_VALIDATION'] = '1'
import argparse
import sys
from latex2pydata import __version__ as latex2pydata_version
from pygments import __version__ as pygments_version
from .version import __version__
from .command_batch import batch
from .command_clean import clean, clean_messages
from .command_config import config
from .command_delete import delete
from .command_highlight import highlight
from .command_styledef import styledef
from .load_data import load_data
from .messages import Messages




def main(latexmintedwindows=None):
    parser = argparse.ArgumentParser(prog='latexminted', allow_abbrev=False)
    parser.set_defaults(func=lambda **x: parser.print_help())
    if latexmintedwindows is None:
        library_version = ', '.join([
            f'latex2pydata {latex2pydata_version}',
            f'pygments {pygments_version}'
        ])
        version = f'latexminted {__version__} (libraries: {library_version})'
    else:
        library_version = ', '.join([
            f'latexminted {__version__}',
            f'latex2pydata {latex2pydata_version}',
            f'pygments {pygments_version}',
        ])
        version = f'latexmintedwindows {latexmintedwindows} (libraries: {library_version})'
    parser.add_argument('--version', action='version', version=version)
    subparsers = parser.add_subparsers(dest='subparser_name')

    parser_batch = subparsers.add_parser('batch', help='Batch process highlight, styledef, and clean')
    parser_batch.set_defaults(func=batch)
    parser_batch.add_argument('--timestamp', help='LaTeX compile timestamp', required=True)
    parser_batch.add_argument('md5', help=r'MD5 hash based \jobname')

    parser_clean = subparsers.add_parser('clean', help='Clean up temp files and unused cache files')
    parser_clean.set_defaults(func=clean)
    parser_clean.add_argument('--timestamp', help='LaTeX compile timestamp', required=True)
    parser_clean.add_argument('md5', help=r'MD5 hash based \jobname')

    parser_config = subparsers.add_parser('config', help='Detect configuration and save it to file for \\input')
    parser_config.set_defaults(func=config)
    parser_config.add_argument('--timestamp', help='LaTeX compile timestamp', required=True)
    parser_config.add_argument('md5', help=r'MD5 hash based \jobname')

    parser_delete = subparsers.add_parser('delete', help='Delete file (restricted filenames only)')
    parser_delete.set_defaults(func=delete)
    parser_delete.add_argument('file', help='File to delete')

    parser_highlight = subparsers.add_parser('highlight', help='Highlight code and save it to file for \\input')
    parser_highlight.set_defaults(func=highlight)
    parser_highlight.add_argument('--timestamp', help='LaTeX compile timestamp', required=True)
    parser_highlight.add_argument('md5', help=r'MD5 hash based \jobname')

    parser_styledef = subparsers.add_parser('styledef', help='Generate highlighting style definition and save it to file for \\input')
    parser_styledef.set_defaults(func=styledef)
    parser_styledef.add_argument('--timestamp', help='LaTeX compile timestamp', required=True)
    parser_styledef.add_argument('md5', help=r'MD5 hash based \jobname')

    cmdline_args = parser.parse_args()

    func_keys = set(['md5', 'file', 'timestamp'])
    func_args = {k: v for k, v in vars(cmdline_args).items() if k in func_keys}
    md5: str | None = func_args.get('md5')
    timestamp: str | None = func_args.get('timestamp')
    if md5 is None or timestamp is None:
        cmdline_args.func(**func_args)
        sys.exit()

    clean_messages(md5=md5)
    messages = Messages(md5=md5)
    func_args['messages'] = messages
    try:
        data = load_data(md5=md5, messages=messages, timestamp=timestamp, command=cmdline_args.subparser_name)
    except Exception as e:
        messages.append_error(
            rf'Failed due to unexpected error (see \detokenize{{"{messages.errlog_file_name}"}} if it exists)'
        )
        messages.append_errlog(e)
        messages.communicate()
        sys.exit(1)

    if messages.data_file_not_found and cmdline_args.subparser_name == 'config':
        config(**func_args)
        sys.exit(1)
    if messages.has_errors():
        messages.communicate()
        sys.exit(1)
    if data is None:
        messages.append_error('Unexpectedly received no data without any error messages')
        messages.communicate()
        sys.exit(1)

    func_args['data'] = data
    cmdline_args.func(**func_args)
    messages.communicate()
    if messages.has_errors():
        sys.exit(1)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

import hashlib
from pygments.lexer import Lexer
from ..err import CustomLexerError, CustomLexerSecurityError, MissingLatexMintedRCError
from ._latexminted_config import load_latexminted_config
from ._restricted_pathlib import RestrictedPath




def load_custom_lexer(lexer: str) -> type[Lexer]:
    lexer_path = RestrictedPath(lexer)
    class_name: str = 'CustomLexer'
    if ':' in lexer_path.name:
        file_name, class_name = lexer_path.name.split(':', 1)
        lexer_path = lexer_path.parent / file_name
    if not lexer_path.name.endswith('.py'):
        raise CustomLexerError('Custom lexers must be in *.py files')

    latexminted_config_data = load_latexminted_config()
    if latexminted_config_data is None:
        raise MissingLatexMintedRCError('Missing or invalid ~/.latexminted_config; custom lexers are disabled')
    if 'custom_lexers' not in latexminted_config_data:
        raise CustomLexerSecurityError('No custom lexers are enabled in ~/.latexminted_config')

    try:
        lexer_bytes = lexer_path.read_bytes()
    except FileNotFoundError:
        raise CustomLexerError(f'Custom lexer "{lexer_path.name}" was not found')
    except PermissionError:
        raise CustomLexerError(f'Insufficient permission to open custom lexer "{lexer_path.name}"')

    hasher = hashlib.sha256()
    hasher.update(lexer_bytes)
    hash = hasher.hexdigest()
    if latexminted_config_data['custom_lexers'].get(lexer_path.name, '').lower() != hash:
        raise CustomLexerSecurityError(
            f'Custom lexer "{lexer_path.name}" is not enabled in ~/.latexminted_config; check that SHA-256 hash is present and updated'
        )

    namespace = {}
    try:
        exec(lexer_bytes, namespace)
    except Exception:
        raise CustomLexerError(f'Failed to exec custom lexer "{lexer}"; check the lexer file for errors')
    if class_name not in namespace:
        raise CustomLexerError(f'Custom lexer class "{class_name}" was not created by "{lexer_path.name}"')
    lexer_class = namespace[class_name]
    if not issubclass(lexer_class, Lexer):
        raise CustomLexerError(f'Custom lexer class "{class_name}" is not a subclass of Pygments Lexer class')
    return lexer_class



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

import textwrap
import traceback
from typing import Any
from .restricted import cwd_path, RestrictedPath, tempfiledir_path, texmfoutput_path




class Messages(object):
    def __init__(self, *, md5: str):
        self._md5 = md5
        self._warnings: list[str] = []
        self._errors: list[str] = []
        self._errlogs: list[str] = []
        self.message_file_name: str = f'_{self._md5}.message.minted'
        self.errlog_file_name: str = f'_{self._md5}.errlog.minted'
        self.data_file_not_found: bool = False
        self._jobname: str | None = None
        self._currentfilepath: str | None = None
        self._currentfile: str | None = None
        self._inputlineno: str | None = None


    def set_context(self, data: dict[str, Any] | None = None):
        if data is None:
            self._jobname = None
            self._currentfilepath = None
            self._currentfile = None
            self._inputlineno = None
        else:
            self._jobname = data['jobname']
            self._currentfilepath = data['currentfilepath']
            self._currentfile = data['currentfile']
            self._inputlineno = data['inputlineno']

    def _add_context(self, message: str) -> str:
        if self._jobname is None:
            return message

        if self._currentfilepath:
            path = self._currentfilepath
            if not RestrictedPath(path).is_absolute() and not path.startswith('.'):
                path = f'./{path}'
            if not path.endswith('/'):
                path += '/'
        else:
            path = './'
        if self._currentfile:
            name = self._currentfile
        else:
            name = self._jobname
            if (name.startswith('"') and name.endswith('"')) or (name.startswith("'") and name.endswith("'")):
                name = name[1:-1]
        number = self._inputlineno
        return rf'^^J => \detokenize{{{path}{name}:{number}:}} {message}'


    def append_warning(self, message: str):
        self._warnings.append(self._add_context(message))

    def append_error(self, message: str):
        self._errors.append(self._add_context(message))

    def append_errlog(self, message: str | Exception):
        if isinstance(message, str):
            self._errlogs.append(message)
        elif isinstance(message, Exception):
            self._errlogs.append(
                textwrap.dedent(''.join(traceback.format_tb(message.__traceback__)))
            )
        else:
            raise TypeError

    def has_errors(self) -> bool:
        return len(self._errors) > 0


    def communicate(self):
        if not self._warnings and not self._errors and not self._errlogs:
            return

        message_lines = []
        if self._warnings:
            message_lines.append(r'\def\minted@exec@warning{%')
            for message in self._warnings:
                message_lines.append(rf'  \minted@warning{{{message}}}%')
            message_lines.append('}%')
        if self._errors:
            message_lines.append(r'\def\minted@exec@error{%')
            for message in self._errors:
                message_lines.append(rf'  \minted@error{{{message}}}%')
            message_lines.append(r'}%')

        write_paths: list[RestrictedPath] = [tempfiledir_path]
        if cwd_path is not tempfiledir_path:
            write_paths.append(cwd_path)
        if texmfoutput_path is not None and texmfoutput_path not in write_paths:
            write_paths.append(texmfoutput_path)

        if message_lines:
            for write_path in write_paths:
                try:
                    (write_path / self.message_file_name).write_text('\n'.join(message_lines))
                except PermissionError:
                    continue
                else:
                    break

        if self._errlogs:
            for write_path in write_paths:
                try:
                    (write_path / self.errlog_file_name).write_text('\n'.join(self._errlogs))
                except PermissionError:
                    continue
                else:
                    break



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

from typing import Any
from ast import literal_eval
from json import loads as json_loads
try:
    from tomllib import loads as toml_loads
except ImportError:
    toml_loads = None

from ._restricted_pathlib import latexminted_config_read_bytes




_cache = {}

def load_latexminted_config() -> dict[str, Any] | None:
    if 'data' in _cache:
        return _cache['data']

    try:
        latexminted_config_bytes = latexminted_config_read_bytes()
    except (FileNotFoundError, PermissionError):
        _cache['data'] = None
        return _cache['data']

    try:
        latexminted_config_str = latexminted_config_bytes.decode('utf-8-sig')
    except UnicodeDecodeError:
        _cache['data'] = None
        return _cache['data']

    loaders = [json_loads, literal_eval]
    if toml_loads is not None:
        loaders.append(toml_loads)
    data: dict[str, Any] | None = None
    for loader in loaders:
        try:
            data = loader(latexminted_config_str)
        except Exception:
            continue
        else:
            if not isinstance(data, dict) or not all(isinstance(k, str) for k in data):
                data = None
            if data is not None and 'custom_lexers' in data:
                if not all(isinstance(k, str) and isinstance(v, str) for k, v in data['custom_lexers'].items()):
                    data = None
            break
    _cache['data'] = data
    return _cache['data']



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

from typing import Any
from .messages import Messages
from .restricted import latex2pydata_loads, RestrictedPath, tempfiledir_path




def load_data(*, md5: str, messages: Messages, timestamp: str, command: str) -> list[dict[str, Any]] | dict[str, Any] | None:
    data_file_name: str = f'_{md5}.data.minted'
    data_path: RestrictedPath = tempfiledir_path / data_file_name

    data_text: str
    try:
        data_text = data_path.read_text()
    except FileNotFoundError:
        messages.append_error(rf'Failed to find file \detokenize{{"{data_file_name}"}}')
        messages.data_file_not_found = True
    except PermissionError:
        messages.append_error(rf'Insufficient permission to open file \detokenize{{"{data_file_name}"}}')
    except UnicodeDecodeError:
        messages.append_error(rf'Failed to decode file \detokenize{{"{data_file_name}"}} (expected UTF-8)')
    if messages.has_errors():
        return None

    try:
        data = latex2pydata_loads(data_text, schema={'cachefiles': 'list[str]'}, schema_missing='rawstr')
    except Exception as e:
        messages.append_error(
            rf'Failed to load data from file \detokenize{{"{data_file_name}"}} (see \detokenize{{"{messages.errlog_file_name}"}})'
        )
        messages.append_errlog(e)
        return None
    if isinstance(data, dict):
        if command != data['command']:
            messages.append_error(
                rf'''minted data file \detokenize{{"{data_file_name}"}} is for "{data['command']}", but expected "{command}"'''
            )
            return None
        if timestamp != data['timestamp']:
            messages.append_error(
                rf'minted data file \detokenize{{"{data_file_name}"}} has incorrect timestamp'
            )
            return None
    elif isinstance(data, list):
        if command != 'batch':
            messages.append_error(
                rf'''minted data file \detokenize{{"{data_file_name}"}} is for "batch", but expected "{command}"'''
            )
            return None
        valid_commands = set(['styledef', 'highlight', 'clean'])
        if not all(d['command'] in valid_commands for d in data):
            messages.append_error(
                rf'''minted data file \detokenize{{"{data_file_name}"}} is for "batch", but contains invalid data'''
            )
            return None
        if data and timestamp != data[0]['timestamp']:
            messages.append_error(
                rf'minted data file \detokenize{{"{data_file_name}"}} has incorrect timestamp'
            )
            return None
    else:
        messages.append_error(
            rf'minted data file \detokenize{{"{data_file_name}"}} contains unexpected data type "{type(data)}"'
        )
        return None
    return data



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

import platform
import shutil
import subprocess
from ..err import SubprocessUnapprovedExecutableError, SubprocessExecutableNotFoundError, SubprocessExecutablePathSecurityError
from ._restricted_pathlib import RestrictedPath




approved_executables = set([
    'kpsewhich',
])




def restricted_run(args: list[str]) -> subprocess.CompletedProcess:
    '''
    Run a command securely, consistent with TeX restricted shell escape.

     *  The executable must be in the list of approved executables in this
        file.

     *  The executable must exist on PATH, as found by `shutil.which()`.  It
        must not be in a location writable by LaTeX.  For added security,
        locations writable by LaTeX cannot be under the executable parent
        directory.  Location is checked via `.is_executable_file_location()`
        from `_restricted_pathlib`.

     *  The executable cannot be a batch file (no *.bat or *.cmd):
        https://docs.python.org/3/library/subprocess.html#security-considerations.
        This is enforced in a somewhat redundant fashion by requiring *.exe
        under Windows and completely prohibiting *.bat and *.cmd everywhere.

     *  The subprocess must run with `shell=False`.
    '''

    if not isinstance(args, list) or not args or not all(isinstance(x, str) for x in args):
        raise TypeError

    executable = args[0]
    if executable not in approved_executables:
        raise SubprocessUnapprovedExecutableError(f'Executable "{executable}" is not in the approved list')

    which_executable = shutil.which(executable)
    if not which_executable:
        raise SubprocessExecutableNotFoundError(f'Executable "{executable}" was not found')

    which_executable_path = RestrictedPath(which_executable).resolve()
    if platform.system() == 'Windows' and not which_executable_path.name.lower().endswith('.exe'):
        raise SubprocessUnapprovedExecutableError(
            f'Executable "{executable}" resolved to "{which_executable_path.as_posix()}", but *.exe is required'
        )
    if any(which_executable_path.name.lower().endswith(ext) for ext in ('.bat', '.cmd')):
        # This should be redundant
        raise SubprocessUnapprovedExecutableError(f'Executable "{executable}" not allowed (no *.bat or *.cmd)')
    if not which_executable_path.is_executable_file_location():
        raise SubprocessExecutablePathSecurityError(
            f'Executable "{executable}" is located under the current directory, $TEXMFOUTPUT, '
            'or $TEXMF_OUTPUT_DIRECTORY, or one of these locations is under the same directory '
            'as the executable'
        )

    # Use resolved executable path for subprocess.  Using just the executable
    # name should give the same results, but this guards against any mismatch
    # between how Python resolves the executable name with no path versus how
    # the subprocess would resolve it.
    resolved_args = [which_executable_path.as_posix()] + args[1:]
    proc = subprocess.run(resolved_args, shell=False, capture_output=True)
    return proc



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

from pygments.formatters import LatexFormatter
from pygments.styles import get_style_by_name
from pygments.util import ClassNotFound
from .err import PathSecurityError
from .messages import Messages
from .restricted import RestrictedPath




def styledef(*, md5: str, timestamp: str, messages: Messages, data: dict[str, str]) -> str | None:
    style = data['style']
    try:
        StyleClass = get_style_by_name(style)
    except ClassNotFound:
        messages.append_error(rf'Pygments style \detokenize{{"{style}"}} was not found')
        return

    style_defs = LatexFormatter(style=StyleClass, commandprefix=data['commandprefix']).get_style_defs().lstrip()
    styledef_path = RestrictedPath(data['cachepath']) / data['styledeffilename']
    try:
        styledef_path.parent.mkdir(parents=True, exist_ok=True)
        styledef_path.write_text(style_defs)
    except PermissionError:
        messages.append_error(rf'Insufficient permission to write style file for \detokenize{{"{style}"}}')
        return
    except PathSecurityError:
        messages.append_error(
            rf'Cannot write \detokenize{{"{style}"}} style file outside working directory, \detokenize{{TEXMFOUTPUT}}, and \detokenize{{TEXMF_OUTPUT_DIRECTORY}}'
        )
        return
    return data['styledeffilename']



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

from typing import Any
from .command_styledef import styledef
from .command_highlight import highlight
from .command_clean import clean
from .messages import Messages




def batch(*, md5: str, timestamp: str, messages: Messages, data: list[dict[str, Any]]):
    new_cache_file_names: list[str] = []
    for d in data:
        command = d['command']
        if command == 'styledef':
            messages.set_context(d)
            f = styledef(md5=md5, timestamp=timestamp, messages=messages, data=d)
            if f is not None:
                new_cache_file_names.append(f)
        elif command == 'highlight':
            messages.set_context(d)
            f = highlight(md5=md5, timestamp=timestamp, messages=messages, data=d)
            if f is not None:
                new_cache_file_names.append(f)
        elif command == 'clean':
            messages.set_context()
            # Don't need to check whether clean is at the end of the list of
            # commands, since the LaTeX side disables the Python executable
            # immediately after clean.
            clean(md5=md5, timestamp=timestamp, messages=messages, data=d, additional_cache_file_names=new_cache_file_names)
        else:
            raise ValueError

    if data and data[-1]['command'] != 'clean':
        messages.set_context()
        # Batch mode is a special case for clean.  When a document without an
        # existing cache is first compiled, an explicit clean command is not
        # written to the data file because there are no cache files in use
        # since they don't yet exist.  However, at the end of the batch, these
        # files do exist, so clean should be invoked to create an index.
        clean_data = {
            'jobname': data[-1]['jobname'],
            'cachepath': data[-1]['cachepath'],
            'cachefiles': [],
        }
        clean(md5=md5, timestamp=timestamp, messages=messages, data=clean_data, additional_cache_file_names=new_cache_file_names)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations




class LatexMintedError(Exception):
    pass

class SecurityError(LatexMintedError):
    pass

class PathSecurityError(SecurityError):
    pass

class SubprocessError(LatexMintedError):
    pass

class SubprocessUnapprovedExecutableError(SecurityError, SubprocessError):
    pass

class SubprocessExecutableNotFoundError(SubprocessError):
    pass

class SubprocessExecutablePathSecurityError(SecurityError, SubprocessError):
    pass

class MissingLatexMintedRCError(LatexMintedError):
    pass

class CustomLexerError(LatexMintedError):
    pass

class CustomLexerSecurityError(SecurityError, CustomLexerError):
    pass



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

import os
import pathlib
import re
import sys
from ..err import PathSecurityError




# The `type(...)` is needed to inherit the `_flavour` attribute
class AnyPath(type(pathlib.Path())):
    if sys.version_info[:2] < (3, 9):
        def is_relative_to(self, other: AnyPath) -> bool:
            try:
                self.relative_to(other)
            except ValueError:
                return False
            return True



class RestrictedPath(type(AnyPath())):
    '''
    Subclass of `pathlib.Path` (which is system-dependent) that restricts file
    operations to be consistent with TeX restricted shell escape security
    requirements.  For this to be effective, all file operations must go
    through this class; it must not be bypassed with other modules such as
    `os` and `shutil` or functions such as `open()`.

     *  Reading:  Restricted to `.read_text()`, `.read_bytes()`, and
        `.open()`.

     *  Writing:  Restricted to `.write_text()` and `.open()`.  Restricted to
        files under the current working directory, $TEXMFOUTPUT, and
        $TEXMF_OUTPUT_DIRECTORY that have names matching the regular
        expression for minted temp files or cache files.

     *  Deleting files:  Restricted to `.unlink()`.  Restricted to files under
        the current working directory, $TEXMFOUTPUT, and
        $TEXMF_OUTPUT_DIRECTORY that have names matching the regular
        expression for minted temp files or cache files.

     *  Creating directories:  Restricted to `.mkdir()`.  Restricted to
        directories under the current working directory, $TEXMFOUTPUT, and
        $TEXMF_OUTPUT_DIRECTORY.

     *  Deleting directories:  Restricted to `.rmdir()`.  Restricted to
        directories under the current working directory, $TEXMFOUTPUT, and
        $TEXMF_OUTPUT_DIRECTORY.  `super().rmdir()` requires an empty
        directory, so no additional checks are needed to ensure that only
        permitted files are deleted.

     *  Completely prohibited:
         -  `.chmod()`
         -  `.lchmod()`
         -  `.write_bytes()` (use `.write_text()`)
         -  `.rename()`
         -  `.replace()`
         -  `.symlink_to()`
         -  `.hardlink_to()`
         -  `.touch()`
    '''

    # `super().resolve()` is used frequently in determining whether paths are
    # readable/writable/executable.  `.resolve()` and `.parent()` cache and
    # track resolved paths to minimize file system access.
    _resolved_set: set[RestrictedPath] = set()
    _resolve_cache: dict[RestrictedPath, RestrictedPath] = {}

    def resolve(self) -> RestrictedPath:
        try:
            resolved = self._resolve_cache[self]
        except KeyError:
            resolved = super().resolve()
            self._resolved_set.add(resolved)
            self._resolve_cache[self] = resolved
            self._resolve_cache[resolved] = resolved
        return resolved

    @property
    def parent(self) -> RestrictedPath:
        parent = super().parent
        if self in self._resolved_set:
            self._resolved_set.add(parent)
            self._resolve_cache[parent] = parent
        return parent

    # There are currently no restrictions on reading locations, but the
    # implementation allows this to be added.  This is equivalent to
    # TeX Live's `openin_any = a`; see
    # https://tug.org/svn/texlive/trunk/Build/source/texk/kpathsea/texmf.cnf?revision=70942&view=markup#l634.
    _fs_read_roots: set[AnyPath] = set()
    # Similarly, there are no restrictions on reading dotfiles, but the
    # implementation allows this to be added.
    _fs_read_dotfiles: bool = True

    _fs_write_roots: set[AnyPath] = set()
    _fs_write_roots.add(AnyPath.cwd().resolve())
    for variable in ('TEXMFOUTPUT', 'TEXMF_OUTPUT_DIRECTORY'):
        value = os.getenv(variable)
        if value:
            value_path = AnyPath(value).resolve()
            _fs_write_roots.add(value_path)

    # Track readable/writable directories and files separately, since some of
    # the requirements are different.

    _checked_readable_dir_set: set[RestrictedPath] = set()
    _is_readable_dir_set: set[RestrictedPath] = set()

    _checked_readable_file_set: set[RestrictedPath] = set()
    _is_readable_file_set: set[RestrictedPath] = set()

    _checked_writable_dir_set: set[RestrictedPath] = set()
    _is_writable_dir_set: set[RestrictedPath] = set()

    _checked_writable_file_set: set[RestrictedPath] = set()
    _is_writable_file_set: set[RestrictedPath] = set()

    # `[0-9a-zA-Z_-]+` covers temp file names `_<MD5 hash>` plus code cache
    # file names `<MD5 hash>` plus style names `<style name>`.
    _writable_filename_re = re.compile(r'[0-9a-zA-Z_-]+\.(?:config|data|errlog|highlight|index|message|style)\.minted')

    _checked_executable_file_location_set: set[RestrictedPath] = set()
    _is_executable_file_location_set: set[RestrictedPath] = set()

    def is_readable_dir(self) -> bool:
        if self not in self._checked_readable_dir_set:
            resolved = self.resolve()
            if not self._fs_read_roots or any(resolved.is_relative_to(p) for p in self._fs_read_roots):
                self._is_readable_dir_set.add(self)
            self._checked_readable_dir_set.add(self)
        return self in self._is_readable_dir_set

    def is_readable_file(self) -> bool:
        if self not in self._checked_readable_file_set:
            resolved = self.resolve()
            if (resolved.parent.is_readable_dir() and (self._fs_read_dotfiles or not resolved.name.startswith('.'))):
                self._is_readable_file_set.add(self)
            self._checked_readable_file_set.add(self)
        return self in self._is_readable_file_set

    def is_writable_dir(self) -> bool:
        if self not in self._checked_writable_dir_set:
            resolved = self.resolve()
            if any(resolved.is_relative_to(p) for p in self._fs_write_roots):
                self._is_writable_dir_set.add(self)
            self._checked_writable_dir_set.add(self)
        return self in self._is_writable_dir_set

    def is_writable_file(self) -> bool:
        if self not in self._checked_writable_file_set:
            resolved = self.resolve()
            if (resolved.parent.is_writable_dir() and self._writable_filename_re.fullmatch(resolved.name)):
                self._is_writable_file_set.add(self)
            self._checked_writable_file_set.add(self)
        return self in self._is_writable_file_set

    def is_executable_file_location(self) -> bool:
        '''
        Note that this only checks whether the path is outside prohibited or
        problematic locations, hence the `_location`.  It does NOT check
        whether the file is executable or is on PATH.  For correct results,
        the executable should be located with `shutil.which()`, and then the
        path derived from this should be checked with
        `.is_executable_file_location()`.  See
        `restricted_subprocess.restricted_run()`.
        '''
        if self not in self._checked_executable_file_location_set:
            resolved = self.resolve()
            parent = resolved.parent
            if (not any(resolved.is_relative_to(p) for p in self._fs_write_roots) and
                    not any(p.is_relative_to(parent) for p in self._fs_write_roots)):
                self._is_executable_file_location_set.add(self)
            self._checked_executable_file_location_set.add(self)
        return self in self._is_executable_file_location_set


    def chmod(self, *args, **kwargs):
        raise NotImplementedError

    def lchmod(self, *args, **kwargs):
        raise NotImplementedError

    def mkdir(self, *args, **kwargs):
        if not self.is_writable_dir():
            raise PathSecurityError(
                f'Cannot create directory because it is outside permitted locations:  "{self.as_posix()}"'
            )
        return super().mkdir(*args, **kwargs)

    def open(self, mode: str, **kwargs):
        # This check is redundant for the `.read_text()`, `.write_text()`,
        # etc. cases which call `.open()` internally.  However, checks on all
        # methods are needed for completeness, and also to guard against the
        # possibility of future implementation changes in `pathlib`.
        if mode in ('r', 'rb'):
            if not self.is_readable_file():
                raise PathSecurityError(
                    f'Cannot read file because it is outside permitted locations:  "{self.as_posix()}"'
                )
        elif mode in ('w', 'wb'):
            if not self.is_writable_file():
                raise PathSecurityError(
                    f'Cannot write file because it is outside permitted locations:  "{self.as_posix()}"'
                )
        else:
            raise NotImplementedError
        return super().open(mode=mode, **kwargs)

    def read_bytes(self, *args, **kwargs):
        if not self.is_readable_file():
            raise PathSecurityError(
                f'Cannot read file because it is outside permitted locations:  "{self.as_posix()}"'
            )
        return super().read_bytes(*args, **kwargs)

    def read_text(self, encoding='utf-8-sig', errors='strict') -> str:
        if not self.is_readable_file():
            raise PathSecurityError(
                f'Cannot read file because it is outside permitted locations:  "{self.as_posix()}"'
            )
        return super().read_text(encoding=encoding, errors=errors)

    def rename(self, *args, **kwargs):
        raise NotImplementedError

    def replace(self, *args, **kwargs):
        raise NotImplementedError

    def rmdir(self):
        if not self.is_writable_dir():
            raise PathSecurityError(
                f'Cannot delete directory because it is outside permitted locations:  "{self.as_posix()}"'
            )
        return super().rmdir()

    def symlink_to(self, *args, **kwargs):
        raise NotImplementedError

    def hardlink_to(self, *args, **kwargs):
        raise NotImplementedError

    def touch(self, *args, **kwargs):
        raise NotImplementedError

    def unlink(self, missing_ok: bool = False):
        if not self.is_writable_file():
            raise PathSecurityError(
                f'Cannot delete file because it is outside permitted locations:  "{self.as_posix()}"'
            )
        return super().unlink(missing_ok=missing_ok)

    def write_bytes(self, *args, **kwargs):
        raise NotImplementedError

    def write_text(self, data: str, encoding: str = 'utf8', **kwargs):
        if not self.is_writable_file():
            raise PathSecurityError(
                f'Cannot write file because it is outside permitted locations:  "{self.as_posix()}"'
            )
        return super().write_text(data, encoding=encoding, **kwargs)




def latexminted_config_read_bytes() -> bytes:
    '''
    Read the minted config file in the user home directory.  This is a dot
    file and thus cannot be accessed via `RestrictedPath`.
    '''
    return AnyPath('~/.latexminted_config').expanduser().read_bytes()




cwd_path: RestrictedPath = RestrictedPath.cwd().resolve()

# Ignore $TEXMFOUTPUT in determining location to write temp files, since
# it's a fallback if the current directory isn't writable.  $TEXMFOUTPUT is
# only used for communicating error messages.
tempfiledir_path: RestrictedPath
_texmf_output_directory = os.getenv('TEXMF_OUTPUT_DIRECTORY')
if _texmf_output_directory:
    tempfiledir_path = RestrictedPath(_texmf_output_directory).resolve()
else:
    tempfiledir_path = cwd_path
del _texmf_output_directory

texmfoutput_path: RestrictedPath | None = None
_texmfoutput = os.getenv('TEXMFOUTPUT')
if _texmfoutput:
    texmfoutput_path = RestrictedPath(_texmfoutput).resolve()
del _texmfoutput

# Version of `tempfiledir_path` to use within `\input` in LaTeX
input_tempfiledir_path_str: str
try:
    input_tempfiledir_path_str = tempfiledir_path.relative_to(cwd_path).as_posix()
except ValueError:
    input_tempfiledir_path_str = tempfiledir_path.as_posix()
if input_tempfiledir_path_str == '.':
    input_tempfiledir_path_str = ''
elif not input_tempfiledir_path_str.endswith('/'):
    input_tempfiledir_path_str += '/'



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from ._restricted_pathlib import RestrictedPath
from ._restricted_subprocess import restricted_run

from ._stdlib import json_dumps, json_loads, os_environ
from ._restricted_pathlib import cwd_path, input_tempfiledir_path_str, tempfiledir_path, texmfoutput_path
from ._lib import latex2pydata_loads

from ._load_custom_lexer import load_custom_lexer



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

from .err import PathSecurityError
from .restricted import cwd_path, tempfiledir_path




def delete(*, file: str):
    paths = [tempfiledir_path]
    if cwd_path is not tempfiledir_path:
        paths.append(cwd_path)
    for path in paths:
        try:
            (path / file).unlink(missing_ok=True)
        except (PermissionError, PathSecurityError):
            pass



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from json import dumps as json_dumps
from json import loads as json_loads
from  os import environ as os_environ
